"""Tests for extract_labware_definitions()."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Set

import pytest

from opentrons_shared_data import get_shared_data_root
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.protocol_reader import (
    JsonProtocolConfig,
    ProtocolFileRole,
    ProtocolSource,
    ProtocolSourceFile,
    PythonProtocolConfig,
    extract_labware_definitions,
)
from opentrons.protocols.api_support.types import APIVersion


@dataclass(frozen=True)
class _LabwareSummary:
    """Brief information to uniquely identify a labware definition."""

    load_name: str
    display_name: str

    @classmethod
    def from_full_definition(
        cls, full_definition: LabwareDefinition
    ) -> _LabwareSummary:
        return cls(
            load_name=full_definition.parameters.loadName,
            display_name=full_definition.metadata.displayName,
        )


@dataclass(frozen=True)
class _JsonProtocolFixture:
    """Information about a JSON protocol fixture in shared-data."""

    schema_version: int
    file_name: str
    expected_labware: Set[_LabwareSummary]

    @property
    def path(self) -> Path:
        return (
            get_shared_data_root()
            / "protocol"
            / "fixtures"
            / str(self.schema_version)
            / self.file_name
        )


def _get_standard_labware_path(load_name: str) -> Path:
    return (
        get_shared_data_root() / "labware" / "definitions" / "2" / load_name / "1.json"
    )


json_protocol_fixtures = [
    _JsonProtocolFixture(
        schema_version=3,
        file_name="simple.json",
        expected_labware={
            _LabwareSummary(
                load_name="opentrons_1_trash_1100ml_fixed",
                display_name="Opentrons Fixed Trash",
            ),
            _LabwareSummary(
                load_name="opentrons_96_tiprack_10ul",
                display_name="Opentrons 96 Tip Rack 10 µL",
            ),
            _LabwareSummary(
                load_name="foo_8_plate_33ul",
                display_name="Foo 8 Well Plate 33uL",
            ),
        },
    ),
    _JsonProtocolFixture(
        schema_version=4,
        file_name="simpleV4.json",
        expected_labware={
            _LabwareSummary(
                load_name="opentrons_1_trash_1100ml_fixed",
                display_name="Opentrons Fixed Trash",
            ),
            _LabwareSummary(
                load_name="opentrons_96_tiprack_10ul",
                display_name="Opentrons 96 Tip Rack 10 µL",
            ),
            _LabwareSummary(
                load_name="foo_8_plate_33ul",
                display_name="Foo 8 Well Plate 33uL",
            ),
        },
    ),
    _JsonProtocolFixture(
        schema_version=5,
        file_name="simpleV5.json",
        expected_labware={
            _LabwareSummary(
                load_name="opentrons_1_trash_1100ml_fixed",
                display_name="Opentrons Fixed Trash",
            ),
            _LabwareSummary(
                load_name="opentrons_96_tiprack_10ul",
                display_name="Opentrons 96 Tip Rack 10 µL",
            ),
            _LabwareSummary(
                load_name="foo_8_plate_33ul",
                display_name="Foo 8 Well Plate 33uL",
            ),
        },
    ),
    _JsonProtocolFixture(
        schema_version=6,
        file_name="simpleV6.json",
        expected_labware={
            _LabwareSummary(
                load_name="opentrons_1_trash_1100ml_fixed",
                display_name="Opentrons Fixed Trash",
            ),
            _LabwareSummary(
                load_name="opentrons_96_tiprack_10ul",
                display_name="Opentrons 96 Tip Rack 10 µL",
            ),
            _LabwareSummary(
                load_name="foo_8_plate_33ul",
                display_name="Foo 8 Well Plate 33uL",
            ),
        },
    ),
    _JsonProtocolFixture(
        schema_version=7,
        file_name="simpleV7.json",
        expected_labware={
            _LabwareSummary(
                load_name="opentrons_1_trash_1100ml_fixed",
                display_name="Opentrons Fixed Trash",
            ),
            _LabwareSummary(
                load_name="opentrons_96_tiprack_10ul",
                display_name="Opentrons 96 Tip Rack 10 µL",
            ),
            _LabwareSummary(
                load_name="foo_8_plate_33ul",
                display_name="Foo 8 Well Plate 33uL",
            ),
        },
    ),
]


@pytest.mark.parametrize("json_protocol", json_protocol_fixtures)
async def test_extraction_from_json_protocol(
    json_protocol: _JsonProtocolFixture,
) -> None:
    """Test extraction of labware definitions from JSON protocols of various versions.

    Since `config` is a `JsonProtocolConfig`, it should interpret the `main_file` as a
    JSON protocol and extract all definitions from it.
    """
    protocol_source = ProtocolSource(
        files=[ProtocolSourceFile(path=json_protocol.path, role=ProtocolFileRole.MAIN)],
        config=JsonProtocolConfig(schema_version=json_protocol.schema_version),
        # The subject should ignore these attributes:
        directory=None,
        main_file=json_protocol.path,
        metadata={},
        robot_type="OT-2 Standard",
        content_hash="abc123",
    )
    result = await extract_labware_definitions(protocol_source)
    summarized_result = {
        _LabwareSummary.from_full_definition(labware) for labware in result
    }
    assert summarized_result == json_protocol.expected_labware


async def test_extraction_from_json_protocol_ignores_separate_labware_files() -> None:
    """Test extraction from a JSON protocol that also includes separate labware files.

    The separate labware files should be ignored. Only the ones from the main protocol
    file should be returned.
    """
    json_protocol = json_protocol_fixtures[0]

    extra_labware_load_names = {
        "corning_96_wellplate_360ul_flat",
        "opentrons_24_tuberack_generic_2ml_screwcap",
    }

    main_file = ProtocolSourceFile(path=json_protocol.path, role=ProtocolFileRole.MAIN)
    extra_labware_files = [
        ProtocolSourceFile(
            path=_get_standard_labware_path(load_name=load_name),
            role=ProtocolFileRole.LABWARE,
        )
        for load_name in extra_labware_load_names
    ]
    protocol_source = ProtocolSource(
        files=[main_file] + extra_labware_files,
        config=JsonProtocolConfig(schema_version=json_protocol.schema_version),
        # The subject should ignore these attributes:
        directory=None,
        main_file=json_protocol.path,
        robot_type="OT-2 Standard",
        metadata={},
        content_hash="abc123",
    )

    result = await extract_labware_definitions(protocol_source)
    summarized_result = {
        _LabwareSummary.from_full_definition(labware) for labware in result
    }

    # It should include the definitions from the main file:
    assert summarized_result == json_protocol.expected_labware
    # It should not include any definitions from the separate files:
    assert extra_labware_load_names.isdisjoint(
        {labware.load_name for labware in summarized_result}
    )


async def test_extraction_from_python_protocol() -> None:
    """Test extraction of labware definitions from Python protocols.

    Since `config` is a `PythonProtocolConfig`, it should ignore the `main_file` as
    an opaque Python source, and parse other files as labware definitions.
    """
    labware_file_paths = {
        _get_standard_labware_path(load_name="corning_96_wellplate_360ul_flat"),
        _get_standard_labware_path(
            load_name="opentrons_24_tuberack_generic_2ml_screwcap"
        ),
    }

    expected_labware = {
        _LabwareSummary(
            load_name="corning_96_wellplate_360ul_flat",
            display_name="Corning 96 Well Plate 360 µL Flat",
        ),
        _LabwareSummary(
            load_name="opentrons_24_tuberack_generic_2ml_screwcap",
            display_name="Opentrons 24 Tube Rack with Generic 2 mL Screwcap",
        ),
    }

    protocol_source = ProtocolSource(
        files=[
            ProtocolSourceFile(path=labware_file_path, role=ProtocolFileRole.LABWARE)
            for labware_file_path in labware_file_paths
        ],
        config=PythonProtocolConfig(api_version=APIVersion(9001, 314159265358979)),
        # The subject should ignore these attributes:
        directory=None,
        main_file=Path("/this/file/does/not/exist/and/should/not/matter"),
        robot_type="OT-2 Standard",
        metadata={},
        content_hash="abc123",
    )

    result = await extract_labware_definitions(protocol_source)
    summarized_result = {
        _LabwareSummary.from_full_definition(labware) for labware in result
    }
    assert summarized_result == expected_labware



from typing import Sequence

import pytest

from opentrons.hardware_control.emulation.parser import Command, Parser


@pytest.fixture
def parser() -> Parser:
    return Parser()


@pytest.mark.parametrize(
    argnames=["line", "expected"],
    argvalues=[
        ["", []],
        ["G2", [Command(gcode="G2", body="", params={})]],
        [
            "   M123.32 with some trailing nonesense   ",
            [Command(gcode="M123.32", body="with some trailing nonesense", params={})],
        ],
        [
            "G123 V2 X0",
            [Command(gcode="G123", body="V2 X0", params={"V": 2.0, "X": 0.0})],
        ],
        [
            "dfuversionM12.3G2",
            [
                Command(gcode="dfu", body="", params={}),
                Command(gcode="version", body="", params={}),
                Command(gcode="M12.3", body="", params={}),
                Command(gcode="G2", body="", params={}),
            ],
        ],
        # Substring check. Don't confuse G1 with G123 or G123 with G123.2
        [
            "G123G123.2G1",
            [
                Command(gcode="G123", body="", params={}),
                Command(gcode="G123.2", body="", params={}),
                Command(gcode="G1", body="", params={}),
            ],
        ],
        # Negative numbers, no spaces.
        [
            "M123.2 B132C-1D321.2",
            [
                Command(
                    gcode="M123.2",
                    body="B132C-1D321.2",
                    params={"B": 132, "C": -1, "D": 321.2},
                )
            ],
        ],
        # Just key, no value.
        [
            "M123.2 LX R",
            [
                Command(
                    gcode="M123.2",
                    body="LX R",
                    params={"L": None, "X": None, "R": None},
                )
            ],
        ],
    ],
)
def test_parse_command(parser: Parser, line: str, expected: Sequence[Command]) -> None:
    """It should parse the commands and parameters."""
    result = list(parser.parse(line))

    assert result == expected


@pytest.mark.parametrize(
    argnames=["line"],
    argvalues=[["not a gcode G123"], ["X123 G123"], ["   DFU G123"], ["no gcode"]],
)
def test_fail_parse(parser: Parser, line: str) -> None:
    """It should raise an exception"""
    with pytest.raises(ValueError, match="Invalid content"):
        list(parser.parse(line=line))



"""Types for Tasks."""

import asyncio
from dataclasses import dataclass
from datetime import datetime

from opentrons.protocol_engine.errors import ErrorOccurrence


@dataclass
class _BaseTask:
    """A base task representation."""

    id: str
    createdAt: datetime


@dataclass
class Task(_BaseTask):
    """A task representation."""

    asyncioTask: asyncio.Task[None]


@dataclass
class FinishedTask(_BaseTask):
    """A finished task representation."""

    finishedAt: datetime
    error: ErrorOccurrence | None


@dataclass
class TaskSummary:
    """Task info for use in summary lists."""

    id: str
    createdAt: datetime
    finishedAt: datetime | None = None
    error: ErrorOccurrence | None = None



"""Test move to addressable area for drop tip commands."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import StallOrCollisionDetectedError

from opentrons.protocol_engine import AddressableOffsetVector, DeckPoint
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.move_to_addressable_area_for_drop_tip import (
    MoveToAddressableAreaForDropTipImplementation,
    MoveToAddressableAreaForDropTipParams,
    MoveToAddressableAreaForDropTipResult,
)
from opentrons.protocol_engine.commands.movement_common import StallOrCollisionError
from opentrons.protocol_engine.execution import MovementHandler
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.types import Point


@pytest.fixture
def subject(
    state_view: StateView, movement: MovementHandler, model_utils: ModelUtils
) -> MoveToAddressableAreaForDropTipImplementation:
    """Get a command implementation with injected dependencies."""
    return MoveToAddressableAreaForDropTipImplementation(
        state_view=state_view, movement=movement, model_utils=model_utils
    )


async def test_move_to_addressable_area_for_drop_tip_implementation(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    subject: MoveToAddressableAreaForDropTipImplementation,
) -> None:
    """A MoveToAddressableAreaForDropTip command should have an execution implementation."""
    data = MoveToAddressableAreaForDropTipParams(
        pipetteId="abc",
        addressableAreaName="123",
        offset=AddressableOffsetVector(x=1, y=2, z=3),
        forceDirect=True,
        minimumZHeight=4.56,
        speed=7.89,
        alternateDropLocation=True,
        ignoreTipConfiguration=False,
    )

    decoy.when(
        state_view.geometry.get_next_tip_drop_location_for_addressable_area(
            addressable_area_name="123", pipette_id="abc"
        )
    ).then_return(AddressableOffsetVector(x=10, y=11, z=12))

    decoy.when(
        await movement.move_to_addressable_area(
            pipette_id="abc",
            addressable_area_name="123",
            offset=AddressableOffsetVector(x=10, y=11, z=12),
            force_direct=True,
            minimum_z_height=4.56,
            speed=7.89,
            stay_at_highest_possible_z=False,
            ignore_tip_configuration=False,
            highest_possible_z_extra_offset=None,
        )
    ).then_return(Point(x=9, y=8, z=7))

    result = await subject.execute(data)

    assert result == SuccessData(
        public=MoveToAddressableAreaForDropTipResult(position=DeckPoint(x=9, y=8, z=7)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="abc",
                new_location=update_types.AddressableArea(addressable_area_name="123"),
                new_deck_point=DeckPoint(x=9, y=8, z=7),
            ),
            addressable_area_used=update_types.AddressableAreaUsedUpdate(
                addressable_area_name="123"
            ),
        ),
    )


async def test_move_to_addressable_area_for_drop_tip_handles_stalls(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    model_utils: ModelUtils,
    subject: MoveToAddressableAreaForDropTipImplementation,
) -> None:
    """A MoveToAddressableAreaForDropTip command should have an execution implementation."""
    data = MoveToAddressableAreaForDropTipParams(
        pipetteId="abc",
        addressableAreaName="123",
        offset=AddressableOffsetVector(x=1, y=2, z=3),
        forceDirect=True,
        minimumZHeight=4.56,
        speed=7.89,
        alternateDropLocation=True,
        ignoreTipConfiguration=False,
    )

    decoy.when(
        state_view.geometry.get_next_tip_drop_location_for_addressable_area(
            addressable_area_name="123", pipette_id="abc"
        )
    ).then_return(AddressableOffsetVector(x=10, y=11, z=12))

    decoy.when(
        await movement.move_to_addressable_area(
            pipette_id="abc",
            addressable_area_name="123",
            offset=AddressableOffsetVector(x=10, y=11, z=12),
            force_direct=True,
            minimum_z_height=4.56,
            speed=7.89,
            stay_at_highest_possible_z=False,
            ignore_tip_configuration=False,
            highest_possible_z_extra_offset=None,
        )
    ).then_raise(StallOrCollisionDetectedError())
    timestamp = datetime.now()
    test_id = "test-id"
    decoy.when(model_utils.generate_id()).then_return(test_id)
    decoy.when(model_utils.get_timestamp()).then_return(timestamp)

    result = await subject.execute(data)

    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id=test_id, createdAt=timestamp, wrappedErrors=[matchers.Anything()]
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.CLEAR,
            addressable_area_used=update_types.AddressableAreaUsedUpdate(
                addressable_area_name="123"
            ),
        ),
    )



"""Test Thermocycler run profile command implementation."""

from decoy import Decoy

from opentrons.hardware_control.modules import Thermocycler
from opentrons.protocol_engine.commands import thermocycler as tc_commands
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.thermocycler.run_profile import (
    RunProfileImpl,
)
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.state.module_substates import (
    ThermocyclerModuleId,
    ThermocyclerModuleSubState,
)
from opentrons.protocol_engine.state.state import StateView


async def test_run_profile(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
) -> None:
    """It should be able to execute the specified module's profile run."""
    subject = RunProfileImpl(state_view=state_view, equipment=equipment)

    step_data = [
        tc_commands.RunProfileStepParams(celsius=12.3, holdSeconds=45, rampRate=0.0),
        tc_commands.RunProfileStepParams(celsius=45.6, holdSeconds=78, rampRate=0.0),
    ]
    data = tc_commands.RunProfileParams(
        moduleId="input-thermocycler-id",
        profile=step_data,
        blockMaxVolumeUl=56.7,
    )
    expected_result = tc_commands.RunProfileResult()

    tc_module_substate = decoy.mock(cls=ThermocyclerModuleSubState)
    tc_hardware = decoy.mock(cls=Thermocycler)

    decoy.when(
        state_view.modules.get_thermocycler_module_substate("input-thermocycler-id")
    ).then_return(tc_module_substate)

    decoy.when(tc_module_substate.module_id).then_return(
        ThermocyclerModuleId("thermocycler-id")
    )

    # Stub temperature validation from hs module view
    decoy.when(tc_module_substate.validate_target_block_temperature(12.3)).then_return(
        32.1
    )
    decoy.when(tc_module_substate.validate_target_block_temperature(45.6)).then_return(
        65.4
    )

    # Stub volume validation from hs module view
    decoy.when(tc_module_substate.validate_max_block_volume(56.7)).then_return(76.5)

    decoy.when(tc_module_substate.validate_ramp_rate(0.0, 12.3)).then_return(0.0)
    decoy.when(tc_module_substate.validate_ramp_rate(0.0, 45.6)).then_return(0.0)

    # Get attached hardware modules
    decoy.when(
        equipment.get_module_hardware_api(ThermocyclerModuleId("thermocycler-id"))
    ).then_return(tc_hardware)

    result = await subject.execute(data)

    decoy.verify(
        await tc_hardware.cycle_temperatures(
            steps=[
                {"temperature": 32.1, "hold_time_seconds": 45, "ramp_rate": 0.0},
                {"temperature": 65.4, "hold_time_seconds": 78, "ramp_rate": 0.0},
            ],
            repetitions=1,
            volume=76.5,
        ),
        times=1,
    )
    assert result == SuccessData(public=expected_result)



"""MovementHandler command subject."""

from typing import NamedTuple, Optional

import pytest
from decoy import Decoy

from opentrons_shared_data.errors.exceptions import PositionUnknownError

from opentrons.hardware_control import API as HardwareAPI
from opentrons.hardware_control.types import CriticalPoint
from opentrons.motion_planning import Waypoint
from opentrons.protocol_engine.execution.equipment import EquipmentHandler
from opentrons.protocol_engine.execution.gantry_mover import GantryMover
from opentrons.protocol_engine.execution.heater_shaker_movement_flagger import (
    HeaterShakerMovementFlagger,
)
from opentrons.protocol_engine.execution.movement import MovementHandler
from opentrons.protocol_engine.execution.thermocycler_movement_flagger import (
    ThermocyclerMovementFlagger,
)
from opentrons.protocol_engine.state.motion import PipetteLocationData
from opentrons.protocol_engine.state.state import (
    StateStore,
)
from opentrons.protocol_engine.types import (
    AddressableOffsetVector,
    CurrentWell,
    DeckPoint,
    DeckSlotLocation,
    MotorAxis,
    MovementAxis,
    WellLocation,
    WellOffset,
    WellOrigin,
)
from opentrons.types import DeckSlotName, Mount, MountType, Point


@pytest.fixture
def hardware_api(decoy: Decoy) -> HardwareAPI:
    """Get a mock in the shape of a HardwareAPI."""
    return decoy.mock(cls=HardwareAPI)


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mock in the shape of a StateStore."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def thermocycler_movement_flagger(decoy: Decoy) -> ThermocyclerMovementFlagger:
    """Get a mock in the shape of a ThermocyclerMovementFlagger."""
    return decoy.mock(cls=ThermocyclerMovementFlagger)


@pytest.fixture
def heater_shaker_movement_flagger(decoy: Decoy) -> HeaterShakerMovementFlagger:
    """Get a mock in the shape of a HeaterShakerMovementFlagger."""
    return decoy.mock(cls=HeaterShakerMovementFlagger)


@pytest.fixture
def mock_gantry_mover(decoy: Decoy) -> GantryMover:
    """Get a mock in the shape of a GantryMover."""
    return decoy.mock(cls=GantryMover)


@pytest.fixture
def mock_equipment_handler(decoy: Decoy) -> EquipmentHandler:
    """Get a mock in the shape of an EquipmentHandler."""
    return decoy.mock(cls=EquipmentHandler)


@pytest.fixture
def subject(
    state_store: StateStore,
    hardware_api: HardwareAPI,
    thermocycler_movement_flagger: ThermocyclerMovementFlagger,
    heater_shaker_movement_flagger: HeaterShakerMovementFlagger,
    mock_gantry_mover: GantryMover,
    mock_equipment_handler: EquipmentHandler,
) -> MovementHandler:
    """Create a MovementHandler with its dependencies mocked out."""
    return MovementHandler(
        state_store=state_store,
        hardware_api=hardware_api,
        thermocycler_movement_flagger=thermocycler_movement_flagger,
        heater_shaker_movement_flagger=heater_shaker_movement_flagger,
        gantry_mover=mock_gantry_mover,
        equipment=mock_equipment_handler,
    )


async def test_move_to_well(
    decoy: Decoy,
    state_store: StateStore,
    thermocycler_movement_flagger: ThermocyclerMovementFlagger,
    heater_shaker_movement_flagger: HeaterShakerMovementFlagger,
    mock_gantry_mover: GantryMover,
    subject: MovementHandler,
) -> None:
    """Move requests should call hardware controller with movement data."""
    well_location = WellLocation(
        origin=WellOrigin.BOTTOM,
        offset=WellOffset(x=0, y=0, z=1),
    )
    decoy.when(state_store.labware.get_location(labware_id="labware-id")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    )
    decoy.when(
        state_store.modules.get_heater_shaker_movement_restrictors()
    ).then_return([])

    decoy.when(state_store.geometry.get_ancestor_slot_name("labware-id")).then_return(
        DeckSlotName.SLOT_1
    )

    decoy.when(state_store.pipettes.get_channels("pipette-id")).then_return(1)
    decoy.when(state_store.labware.is_tiprack("labware-id")).then_return(False)

    decoy.when(
        state_store.motion.get_pipette_location(
            pipette_id="pipette-id",
            current_location=None,
        )
    ).then_return(
        PipetteLocationData(
            mount=MountType.LEFT,
            critical_point=CriticalPoint.FRONT_NOZZLE,
        )
    )

    decoy.when(
        await mock_gantry_mover.get_position(
            pipette_id="pipette-id",
        )
    ).then_return(Point(1, 1, 1))

    decoy.when(mock_gantry_mover.get_max_travel_z(pipette_id="pipette-id")).then_return(
        42.0
    )

    decoy.when(
        state_store.pipettes.get_movement_speed(
            pipette_id="pipette-id", requested_speed=45.6
        )
    ).then_return(39339.5)

    decoy.when(
        state_store.motion.get_movement_waypoints_to_well(
            origin=Point(1, 1, 1),
            origin_cp=CriticalPoint.FRONT_NOZZLE,
            max_travel_z=42.0,
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="B2",
            well_location=well_location,
            current_well=None,
            force_direct=True,
            minimum_z_height=12.3,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(
        [Waypoint(Point(1, 2, 3), CriticalPoint.XY_CENTER), Waypoint(Point(4, 5, 6))]
    )

    decoy.when(
        await mock_gantry_mover.move_to(
            pipette_id="pipette-id",
            waypoints=[
                Waypoint(Point(1, 2, 3), CriticalPoint.XY_CENTER),
                Waypoint(Point(4, 5, 6)),
            ],
            speed=39339.5,
        ),
    ).then_return(Point(4, 5, 6))

    result = await subject.move_to_well(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="B2",
        well_location=well_location,
        force_direct=True,
        minimum_z_height=12.3,
        speed=45.6,
    )

    assert result == Point(x=4, y=5, z=6)

    decoy.verify(
        await thermocycler_movement_flagger.ensure_labware_in_open_thermocycler(
            labware_parent=DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
        ),
        heater_shaker_movement_flagger.raise_if_movement_restricted(
            hs_movement_restrictors=[],
            destination_slot=1,
            is_multi_channel=False,
            destination_is_tip_rack=False,
        ),
    )


async def test_move_to_well_from_starting_location(
    decoy: Decoy,
    state_store: StateStore,
    thermocycler_movement_flagger: ThermocyclerMovementFlagger,
    heater_shaker_movement_flagger: HeaterShakerMovementFlagger,
    mock_gantry_mover: GantryMover,
    subject: MovementHandler,
) -> None:
    """It should be able to move to a well from a start location."""
    well_location = WellLocation(
        origin=WellOrigin.BOTTOM,
        offset=WellOffset(x=0, y=0, z=1),
    )

    current_well = CurrentWell(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="B2",
    )
    decoy.when(state_store.labware.get_location(labware_id="labware-id")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    )
    decoy.when(
        state_store.modules.get_heater_shaker_movement_restrictors()
    ).then_return([])

    decoy.when(state_store.geometry.get_ancestor_slot_name("labware-id")).then_return(
        DeckSlotName.SLOT_1
    )

    decoy.when(state_store.pipettes.get_channels("pipette-id")).then_return(1)
    decoy.when(state_store.labware.is_tiprack("labware-id")).then_return(False)

    decoy.when(
        state_store.motion.get_pipette_location(
            pipette_id="pipette-id",
            current_location=current_well,
        )
    ).then_return(
        PipetteLocationData(
            mount=MountType.RIGHT,
            critical_point=CriticalPoint.XY_CENTER,
        )
    )

    decoy.when(
        await mock_gantry_mover.get_position(
            pipette_id="pipette-id",
        )
    ).then_return(Point(1, 2, 5))

    decoy.when(mock_gantry_mover.get_max_travel_z(pipette_id="pipette-id")).then_return(
        42.0
    )

    decoy.when(
        state_store.motion.get_movement_waypoints_to_well(
            current_well=current_well,
            origin=Point(1, 2, 5),
            origin_cp=CriticalPoint.XY_CENTER,
            max_travel_z=42.0,
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="B2",
            well_location=well_location,
            force_direct=False,
            minimum_z_height=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return([Waypoint(Point(1, 2, 3), CriticalPoint.XY_CENTER)])

    decoy.when(
        state_store.pipettes.get_movement_speed(
            pipette_id="pipette-id", requested_speed=None
        )
    ).then_return(39339.5)

    decoy.when(
        await mock_gantry_mover.move_to(
            pipette_id="pipette-id",
            waypoints=[Waypoint(Point(1, 2, 3), CriticalPoint.XY_CENTER)],
            speed=39339.5,
        ),
    ).then_return(Point(4, 5, 6))

    result = await subject.move_to_well(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="B2",
        well_location=well_location,
        current_well=current_well,
    )

    assert result == Point(4, 5, 6)

    decoy.verify(
        await thermocycler_movement_flagger.ensure_labware_in_open_thermocycler(
            labware_parent=DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
        ),
        heater_shaker_movement_flagger.raise_if_movement_restricted(
            hs_movement_restrictors=[],
            destination_slot=1,
            is_multi_channel=False,
            destination_is_tip_rack=False,
        ),
    )


@pytest.mark.parametrize(
    "stay_at_max_z,z_extra_offset",
    [(True, None), (True, 5.0), (False, None), (False, 5.0)],
)
async def test_move_to_addressable_area(
    decoy: Decoy,
    state_store: StateStore,
    thermocycler_movement_flagger: ThermocyclerMovementFlagger,
    heater_shaker_movement_flagger: HeaterShakerMovementFlagger,
    mock_gantry_mover: GantryMover,
    subject: MovementHandler,
    stay_at_max_z: bool,
    z_extra_offset: Optional[float],
) -> None:
    """Move requests should call hardware controller with movement data."""
    decoy.when(
        state_store.modules.get_heater_shaker_movement_restrictors()
    ).then_return([])

    decoy.when(
        state_store.addressable_areas.get_addressable_area_base_slot("area-name")
    ).then_return(DeckSlotName.SLOT_1)

    decoy.when(state_store.pipettes.get_channels("pipette-id")).then_return(1)

    decoy.when(
        state_store.motion.get_pipette_location(
            pipette_id="pipette-id",
            current_location=None,
        )
    ).then_return(
        PipetteLocationData(
            mount=MountType.LEFT,
            critical_point=CriticalPoint.FRONT_NOZZLE,
        )
    )

    decoy.when(
        await mock_gantry_mover.get_position(
            pipette_id="pipette-id",
        )
    ).then_return(Point(1, 1, 1))

    decoy.when(mock_gantry_mover.get_max_travel_z(pipette_id="pipette-id")).then_return(
        42.0
    )

    decoy.when(
        state_store.pipettes.get_movement_speed(
            pipette_id="pipette-id", requested_speed=45.6
        )
    ).then_return(39339.5)

    decoy.when(
        state_store.motion.get_movement_waypoints_to_addressable_area(
            addressable_area_name="area-name",
            offset=AddressableOffsetVector(x=9, y=8, z=7),
            origin=Point(1, 1, 1),
            origin_cp=CriticalPoint.FRONT_NOZZLE,
            max_travel_z=42.0,
            force_direct=True,
            minimum_z_height=12.3,
            stay_at_max_travel_z=stay_at_max_z,
            ignore_tip_configuration=False,
            max_travel_z_extra_margin=z_extra_offset,
        )
    ).then_return(
        [Waypoint(Point(1, 2, 3), CriticalPoint.XY_CENTER), Waypoint(Point(4, 5, 6))]
    )

    decoy.when(
        await mock_gantry_mover.move_to(
            pipette_id="pipette-id",
            waypoints=[
                Waypoint(Point(1, 2, 3), CriticalPoint.XY_CENTER),
                Waypoint(Point(4, 5, 6)),
            ],
            speed=39339.5,
        ),
    ).then_return(Point(4, 5, 6))

    result = await subject.move_to_addressable_area(
        pipette_id="pipette-id",
        addressable_area_name="area-name",
        offset=AddressableOffsetVector(x=9, y=8, z=7),
        force_direct=True,
        minimum_z_height=12.3,
        speed=45.6,
        stay_at_highest_possible_z=stay_at_max_z,
        ignore_tip_configuration=False,
        highest_possible_z_extra_offset=z_extra_offset,
    )

    assert result == Point(x=4, y=5, z=6)

    decoy.verify(
        heater_shaker_movement_flagger.raise_if_movement_restricted(
            hs_movement_restrictors=[],
            destination_slot=1,
            is_multi_channel=False,
            destination_is_tip_rack=False,
        )
    )


class MoveRelativeSpec(NamedTuple):
    """Test data for move_relative."""

    axis: MovementAxis
    expected_delta: Point
    distance: float = 42.0


@pytest.mark.parametrize(
    MoveRelativeSpec._fields,
    [
        MoveRelativeSpec(
            axis=MovementAxis.X,
            expected_delta=Point(x=42.0, y=0, z=0),
        ),
        MoveRelativeSpec(
            axis=MovementAxis.Y,
            expected_delta=Point(x=0, y=42.0, z=0),
        ),
        MoveRelativeSpec(
            axis=MovementAxis.Z,
            expected_delta=Point(x=0, y=0, z=42.0),
        ),
    ],
)
async def test_move_relative(
    decoy: Decoy,
    state_store: StateStore,
    mock_gantry_mover: GantryMover,
    subject: MovementHandler,
    axis: MovementAxis,
    expected_delta: Point,
    distance: float,
) -> None:
    """Test that move_relative triggers a relative move with the HardwareAPI."""
    decoy.when(
        await mock_gantry_mover.move_relative(
            pipette_id="pipette-id",
            delta=expected_delta,
            speed=39339.5,
        )
    ).then_return(Point(x=1, y=2, z=3))

    decoy.when(
        state_store.pipettes.get_movement_speed(pipette_id="pipette-id")
    ).then_return(39339.5)

    result = await subject.move_relative(
        pipette_id="pipette-id",
        axis=axis,
        distance=distance,
    )

    assert result == Point(x=1, y=2, z=3)


async def test_move_to_coordinates(
    decoy: Decoy,
    state_store: StateStore,
    mock_gantry_mover: GantryMover,
    subject: MovementHandler,
) -> None:
    """Test that move_to_coordinates correctly calls api.move_to."""
    current_position = Point(4.44, 5.55, 6.66)
    destination_deck = DeckPoint(x=1.11, y=2.22, z=3.33)
    destination_point = Point(1.11, 2.22, 3.33)

    planned_waypoint_1 = Waypoint(position=Point(3, 1, 4), critical_point=None)
    planned_waypoint_2 = Waypoint(
        position=Point(1, 5, 9), critical_point=CriticalPoint.XY_CENTER
    )

    decoy.when(
        state_store.motion.get_pipette_location(
            pipette_id="pipette-id",
        )
    ).then_return(
        PipetteLocationData(
            mount=MountType.RIGHT,
            critical_point=CriticalPoint.XY_CENTER,
        )
    )

    decoy.when(
        await mock_gantry_mover.get_position(pipette_id="pipette-id")
    ).then_return(current_position)

    decoy.when(mock_gantry_mover.get_max_travel_z(pipette_id="pipette-id")).then_return(
        5678
    )

    decoy.when(
        state_store.motion.get_movement_waypoints_to_coords(
            origin=current_position,
            dest=destination_point,
            max_travel_z=5678,
            direct=True,
            additional_min_travel_z=1234,
        )
    ).then_return([planned_waypoint_1, planned_waypoint_2])

    decoy.when(
        state_store.pipettes.get_movement_speed(
            pipette_id="pipette-id", requested_speed=567
        )
    ).then_return(39339.5)

    decoy.when(
        await mock_gantry_mover.move_to(
            pipette_id="pipette-id",
            waypoints=[planned_waypoint_1, planned_waypoint_2],
            speed=39339.5,
        )
    ).then_return(Point(x=1, y=5, z=9))

    result = await subject.move_to_coordinates(
        pipette_id="pipette-id",
        deck_coordinates=destination_deck,
        direct=True,
        additional_min_travel_z=1234,
        speed=567,
    )

    assert result == Point(x=1, y=5, z=9)


async def test_retract_axis(
    decoy: Decoy,
    state_store: StateStore,
    mock_gantry_mover: GantryMover,
    subject: MovementHandler,
) -> None:
    """It should delegate to gantry mover to retract the specified axis."""
    await subject.retract_axis(axis=MotorAxis.RIGHT_Z)

    decoy.verify(await mock_gantry_mover.retract_axis(MotorAxis.RIGHT_Z), times=1)


async def test_check_valid_position(
    decoy: Decoy,
    subject: MovementHandler,
    hardware_api: HardwareAPI,
) -> None:
    """It should check for an exception to determine if the position is ok."""
    decoy.when(
        await hardware_api.gantry_position(mount=Mount.LEFT, fail_on_not_homed=True)
    ).then_return(Point(0, 0, 0))
    decoy.when(
        await hardware_api.gantry_position(mount=Mount.RIGHT, fail_on_not_homed=True)
    ).then_raise(PositionUnknownError())

    assert await subject.check_for_valid_position(MountType.LEFT)
    assert not await subject.check_for_valid_position(MountType.RIGHT)



"""Tests for `opentrons.execute`."""

from __future__ import annotations

import io
import json
import textwrap
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterator, List, TextIO, cast

import mock
import pytest
from _pytest.fixtures import SubRequest

from opentrons_shared_data import get_shared_data_root, load_shared_data
from opentrons_shared_data.pipette import (
    load_data as load_pipette_data,
)
from opentrons_shared_data.pipette import (
    pipette_load_name_conversions as pipette_load_name,
)
from opentrons_shared_data.pipette.types import PipetteModel

from opentrons import execute, types
from opentrons.hardware_control import Controller, api
from opentrons.protocol_api.core.engine import ENGINE_CORE_API_VERSION
from opentrons.protocol_engine.types import DeckConfigurationType
from opentrons.protocols.api_support.types import APIVersion
from opentrons.util import entrypoint_util

if TYPE_CHECKING:
    from tests.opentrons.conftest import Bundle, Protocol


HERE = Path(__file__).parent


@pytest.fixture(autouse=True)
def clean_up_hw() -> Iterator[None]:
    """Make sure hardware objects are cleaned up."""
    yield
    execute._LIVE_PROTOCOL_ENGINE_CONTEXTS.close()
    if execute._THREAD_MANAGED_HW is not None:
        execute._THREAD_MANAGED_HW.clean_up()
        execute._THREAD_MANAGED_HW = None


@pytest.fixture(params=[APIVersion(2, 0), ENGINE_CORE_API_VERSION])
def api_version(request: SubRequest) -> APIVersion:
    """Return an API version to test with.

    Newer API versions execute through Protocol Engine, and older API versions don't.
    The two codepaths are very different, so we need to test them both.
    """
    return cast(APIVersion, request.param)


@pytest.fixture
def mock_get_attached_instr(  # noqa: D103
    monkeypatch: pytest.MonkeyPatch,
    virtual_smoothie_env: None,
) -> mock.AsyncMock:
    gai_mock = mock.AsyncMock()

    async def dummy_delay(self: Any, duration_s: float) -> None:
        pass

    monkeypatch.setattr(Controller, "get_attached_instruments", gai_mock)
    monkeypatch.setattr(api.API, "delay", dummy_delay)
    gai_mock.return_value = {
        types.Mount.RIGHT: {"model": None, "id": None},
        types.Mount.LEFT: {"model": None, "id": None},
    }
    return gai_mock


@pytest.fixture(autouse=True)
def mock_deck_configuration(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Override the "host device deck config" to be an OT-2's normal deck config."""

    def mock_get_deck_configuration() -> DeckConfigurationType:
        return [
            ("cutout1", "singleStandardSlot", None),
            ("cutout2", "singleStandardSlot", None),
            ("cutout3", "singleStandardSlot", None),
            ("cutout4", "singleStandardSlot", None),
            ("cutout5", "singleStandardSlot", None),
            ("cutout6", "singleStandardSlot", None),
            ("cutout7", "singleStandardSlot", None),
            ("cutout8", "singleStandardSlot", None),
            ("cutout9", "singleStandardSlot", None),
            ("cutout10", "singleStandardSlot", None),
            ("cutout11", "singleStandardSlot", None),
            ("cutout12", "fixedTrashSlot", None),
        ]

    monkeypatch.setattr(
        entrypoint_util, "get_deck_configuration", mock_get_deck_configuration
    )


@pytest.mark.parametrize(
    ("protocol_file", "expected_entries"),
    [
        (
            "testosaur_v2.py",
            [
                "Picking up tip from A1 of Opentrons OT-2 96 Tip Rack 1000 µL on 1",
                "Aspirating 100.0 uL from A1 of Corning 96 Well Plate 360 µL Flat on 2 at 500.0 uL/sec",
                "Dispensing 100.0 uL into B1 of Corning 96 Well Plate 360 µL Flat on 2 at 1000.0 uL/sec",
                "Dropping tip into H12 of Opentrons OT-2 96 Tip Rack 1000 µL on 1",
            ],
        ),
        (
            "testosaur_v2_14.py",
            [
                "Picking up tip from A1 of Opentrons OT-2 96 Tip Rack 1000 µL on slot 1",
                "Aspirating 100.0 uL from A1 of Corning 96 Well Plate 360 µL Flat on slot 2 at 500.0 uL/sec",
                "Dispensing 100.0 uL into B1 of Corning 96 Well Plate 360 µL Flat on slot 2 at 1000.0 uL/sec",
                "Dropping tip into H12 of Opentrons OT-2 96 Tip Rack 1000 µL on slot 1",
            ],
        ),
    ],
)
def test_execute_function_apiv2(
    protocol: Protocol,
    protocol_file: str,
    expected_entries: List[str],
    virtual_smoothie_env: None,
    mock_get_attached_instr: mock.AsyncMock,
) -> None:
    """Test `execute()` with a Python file."""
    converted_model_v15 = pipette_load_name.convert_pipette_model(
        cast(PipetteModel, "p10_single_v1.5")
    )
    converted_model_v1 = pipette_load_name.convert_pipette_model(
        cast(PipetteModel, "p1000_single_v1")
    )

    mock_get_attached_instr.return_value[types.Mount.LEFT] = {
        "config": load_pipette_data.load_definition(
            converted_model_v15.pipette_type,
            converted_model_v15.pipette_channels,
            converted_model_v15.pipette_version,
            converted_model_v15.oem_type,
        ),
        "id": "testid",
    }
    mock_get_attached_instr.return_value[types.Mount.RIGHT] = {
        "config": load_pipette_data.load_definition(
            converted_model_v1.pipette_type,
            converted_model_v1.pipette_channels,
            converted_model_v1.pipette_version,
            converted_model_v1.oem_type,
        ),
        "id": "testid2",
    }
    entries = []

    def emit_runlog(entry: Any) -> None:
        nonlocal entries
        entries.append(entry)

    execute.execute(protocol.filelike, protocol.filename, emit_runlog=emit_runlog)

    assert [
        item["payload"]["text"] for item in entries if item["$"] == "before"
    ] == expected_entries


def test_execute_function_json_v3(
    get_json_protocol_fixture: Callable[[str, str, bool], str],
    virtual_smoothie_env: None,
    mock_get_attached_instr: mock.AsyncMock,
) -> None:
    """Test `execute()` with a JSONv3 file."""
    jp = get_json_protocol_fixture("3", "simple", False)
    filelike = io.StringIO(jp)
    entries = []

    def emit_runlog(entry: Any) -> None:
        nonlocal entries
        entries.append(entry)

    converted_model_v15 = pipette_load_name.convert_pipette_model(
        cast(PipetteModel, "p10_single_v1.5")
    )
    mock_get_attached_instr.return_value[types.Mount.LEFT] = {
        "config": load_pipette_data.load_definition(
            converted_model_v15.pipette_type,
            converted_model_v15.pipette_channels,
            converted_model_v15.pipette_version,
            converted_model_v15.oem_type,
        ),
        "id": "testid",
    }
    execute.execute(filelike, "simple.json", emit_runlog=emit_runlog)
    assert [item["payload"]["text"] for item in entries if item["$"] == "before"] == [
        "Picking up tip from B1 of Opentrons 96 Tip Rack 10 µL on 1",
        "Aspirating 5.0 uL from A1 of Source Plate on 2 at 3.0 uL/sec",
        "Delaying for 0 minutes and 42.0 seconds",
        "Dispensing 4.5 uL into B1 of Dest Plate on 3 at 2.5 uL/sec",
        "Touching tip",
        "Blowing out into B1 of Dest Plate on 3 at 2.0 uL/sec",
        "Moving to 5",
        "Dropping tip into A1 of Trash on 12",
    ]


def test_execute_function_json_v4(
    get_json_protocol_fixture: Callable[[str, str, bool], str],
    virtual_smoothie_env: None,
    mock_get_attached_instr: mock.AsyncMock,
) -> None:
    """Test `execute()` with a JSONv4 file."""
    jp = get_json_protocol_fixture("4", "simpleV4", False)
    filelike = io.StringIO(jp)
    entries = []

    def emit_runlog(entry: Any) -> None:
        nonlocal entries
        entries.append(entry)

    converted_model_v15 = pipette_load_name.convert_pipette_model(
        cast(PipetteModel, "p10_single_v1.5")
    )
    mock_get_attached_instr.return_value[types.Mount.LEFT] = {
        "config": load_pipette_data.load_definition(
            converted_model_v15.pipette_type,
            converted_model_v15.pipette_channels,
            converted_model_v15.pipette_version,
            converted_model_v15.oem_type,
        ),
        "id": "testid",
    }
    execute.execute(filelike, "simple.json", emit_runlog=emit_runlog)
    assert [item["payload"]["text"] for item in entries if item["$"] == "before"] == [
        "Picking up tip from B1 of Opentrons 96 Tip Rack 10 µL on 1",
        "Aspirating 5.0 uL from A1 of Source Plate on 2 at 3.0 uL/sec",
        "Delaying for 0 minutes and 42.0 seconds",
        "Dispensing 4.5 uL into B1 of Dest Plate on 3 at 2.5 uL/sec",
        "Touching tip",
        "Blowing out into B1 of Dest Plate on 3 at 2.0 uL/sec",
        "Moving to 5",
        "Dropping tip into A1 of Trash on 12",
    ]


def test_execute_function_json_v5(
    get_json_protocol_fixture: Callable[[str, str, bool], str],
    virtual_smoothie_env: None,
    mock_get_attached_instr: mock.AsyncMock,
) -> None:
    """Test `execute()` with a JSONv5 file."""
    jp = get_json_protocol_fixture("5", "simpleV5", False)
    filelike = io.StringIO(jp)
    entries = []

    def emit_runlog(entry: Any) -> None:
        nonlocal entries
        entries.append(entry)

    converted_model_v15 = pipette_load_name.convert_pipette_model(
        cast(PipetteModel, "p10_single_v1.5")
    )
    mock_get_attached_instr.return_value[types.Mount.LEFT] = {
        "config": load_pipette_data.load_definition(
            converted_model_v15.pipette_type,
            converted_model_v15.pipette_channels,
            converted_model_v15.pipette_version,
            converted_model_v15.oem_type,
        ),
        "id": "testid",
    }
    execute.execute(filelike, "simple.json", emit_runlog=emit_runlog)
    assert [item["payload"]["text"] for item in entries if item["$"] == "before"] == [
        "Picking up tip from B1 of Opentrons 96 Tip Rack 10 µL on 1",
        "Aspirating 5.0 uL from A1 of Source Plate on 2 at 3.0 uL/sec",
        "Delaying for 0 minutes and 42.0 seconds",
        "Dispensing 4.5 uL into B1 of Dest Plate on 3 at 2.5 uL/sec",
        "Touching tip",
        "Blowing out into B1 of Dest Plate on 3 at 2.0 uL/sec",
        "Moving to 5",
        "Moving to B2 of Dest Plate on 3",
        "Moving to B2 of Dest Plate on 3",
        "Dropping tip into A1 of Trash on 12",
    ]


def test_execute_function_bundle_apiv2(
    get_bundle_fixture: Callable[[str], Bundle],
    virtual_smoothie_env: None,
    mock_get_attached_instr: mock.AsyncMock,
) -> None:
    """Test `execute()` with a .zip bundle."""
    bundle = get_bundle_fixture("simple_bundle")
    entries = []

    def emit_runlog(entry: Any) -> None:
        nonlocal entries
        entries.append(entry)

    converted_model_v15 = pipette_load_name.convert_pipette_model(
        cast(PipetteModel, "p10_single_v1.5")
    )
    mock_get_attached_instr.return_value[types.Mount.LEFT] = {
        "config": load_pipette_data.load_definition(
            converted_model_v15.pipette_type,
            converted_model_v15.pipette_channels,
            converted_model_v15.pipette_version,
            converted_model_v15.oem_type,
        ),
        "id": "testid",
    }
    execute.execute(
        cast(TextIO, bundle["filelike"]),
        "simple_bundle.zip",
        emit_runlog=emit_runlog,
    )
    assert [item["payload"]["text"] for item in entries if item["$"] == "before"] == [
        "Transferring 1.0 from A1 of FAKE example labware on 1 to A4 of FAKE example labware on 1",
        "Picking up tip from A1 of Opentrons OT-2 96 Tip Rack 10 µL on 3",
        "Aspirating 1.0 uL from A1 of FAKE example labware on 1 at 5.0 uL/sec",
        "Dispensing 1.0 uL into A4 of FAKE example labware on 1 at 10.0 uL/sec",
        "Dropping tip into A1 of Opentrons Fixed Trash on 12",
        "Transferring 2.0 from A1 of FAKE example labware on 1 to A4 of FAKE example labware on 1",
        "Picking up tip from B1 of Opentrons OT-2 96 Tip Rack 10 µL on 3",
        "Aspirating 2.0 uL from A1 of FAKE example labware on 1 at 5.0 uL/sec",
        "Dispensing 2.0 uL into A4 of FAKE example labware on 1 at 10.0 uL/sec",
        "Dropping tip into A1 of Opentrons Fixed Trash on 12",
        "Transferring 3.0 from A1 of FAKE example labware on 1 to A4 of FAKE example labware on 1",
        "Picking up tip from C1 of Opentrons OT-2 96 Tip Rack 10 µL on 3",
        "Aspirating 3.0 uL from A1 of FAKE example labware on 1 at 5.0 uL/sec",
        "Dispensing 3.0 uL into A4 of FAKE example labware on 1 at 10.0 uL/sec",
        "Dropping tip into A1 of Opentrons Fixed Trash on 12",
    ]


class TestExecutePythonLabware:
    """Tests for making sure execute() handles custom labware correctly for Python files."""

    LW_DIR = get_shared_data_root() / "labware" / "fixtures" / "2"
    LW_LOAD_NAME = "fixture_12_trough"
    LW_NAMESPACE = "fixture"

    @pytest.fixture(autouse=True)
    def use_virtual_smoothie_env(self, virtual_smoothie_env: None) -> None:
        """Automatically enable the virtual_smoothie_env fixture for every test."""
        pass

    @pytest.fixture
    def protocol_path(self, tmp_path: Path, api_version: APIVersion) -> Path:
        """Return a path to a Python protocol file that loads a custom labware."""
        path = tmp_path / "protocol.py"
        protocol_source = textwrap.dedent(
            f"""\
            metadata = {{"apiLevel": "{api_version}"}}
            def run(protocol):
                protocol.load_labware(
                    load_name="{self.LW_LOAD_NAME}",
                    location=1,
                    namespace="{self.LW_NAMESPACE}",
                )
            """
        )
        path.write_text(protocol_source)
        return path

    @pytest.fixture
    def protocol_name(self, protocol_path: Path) -> str:
        """Return the file name of the Python protocol file."""
        return protocol_path.name

    @pytest.fixture
    def protocol_filelike(self, protocol_path: Path) -> Generator[TextIO, None, None]:
        """Return the Python protocol file opened as a stream."""
        with open(protocol_path) as file:
            yield file

    @staticmethod
    def test_default_no_custom_labware(
        protocol_filelike: TextIO, protocol_name: str
    ) -> None:
        """By default, no custom labware should be available."""
        with pytest.raises(Exception, match="Labware .+ not found"):
            execute.execute(
                protocol_file=protocol_filelike, protocol_name=protocol_name
            )

    def test_custom_labware_paths(
        self, protocol_filelike: TextIO, protocol_name: str
    ) -> None:
        """Providing custom_labware_paths should make those labware available."""
        execute.execute(
            protocol_file=protocol_filelike,
            protocol_name=protocol_name,
            custom_labware_paths=[str(self.LW_DIR)],
        )

    def test_jupyter(
        self,
        protocol_filelike: TextIO,
        protocol_name: str,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Putting labware in the Jupyter directory should make it available."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util, "JUPYTER_NOTEBOOK_LABWARE_DIR", self.LW_DIR
        )
        execute.execute(protocol_file=protocol_filelike, protocol_name=protocol_name)

    @pytest.mark.xfail(
        strict=True, raises=pytest.fail.Exception
    )  # TODO(mm, 2023-07-14): Fix this bug.
    def test_jupyter_override(
        self,
        protocol_filelike: TextIO,
        protocol_name: str,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Passing any custom_labware_paths should prevent searching the Jupyter directory."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util, "JUPYTER_NOTEBOOK_LABWARE_DIR", self.LW_DIR
        )
        with pytest.raises(Exception, match="Labware .+ not found"):
            execute.execute(
                protocol_file=protocol_filelike,
                protocol_name=protocol_name,
                custom_labware_paths=[],
            )

    @staticmethod
    def test_jupyter_not_on_filesystem(
        protocol_filelike: TextIO,
        protocol_name: str,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """It should tolerate the Jupyter labware directory not existing on the filesystem."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util, "JUPYTER_NOTEBOOK_LABWARE_DIR", HERE / "nosuchdirectory"
        )
        with pytest.raises(Exception, match="Labware .+ not found"):
            execute.execute(
                protocol_file=protocol_filelike, protocol_name=protocol_name
            )


class TestGetProtocolAPILabware:
    """Tests for making sure get_protocol_api() handles extra labware correctly."""

    LW_FIXTURE_DIR = Path("labware/fixtures/2")
    LW_LOAD_NAME = "fixture_12_trough"
    LW_NAMESPACE = "fixture"

    @pytest.fixture(autouse=True)
    def use_virtual_smoothie_env(self, virtual_smoothie_env: None) -> None:
        """Automatically enable the virtual_smoothie_env fixture for every test."""
        pass

    def test_default_no_extra_labware(
        self, api_version: APIVersion, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """By default, no extra labware should be available."""
        context = execute.get_protocol_api(api_version)
        with pytest.raises(Exception, match="Labware .+ not found"):
            context.load_labware(
                load_name=self.LW_LOAD_NAME, location=1, namespace=self.LW_NAMESPACE
            )

    def test_extra_labware(self, api_version: APIVersion) -> None:
        """Providing extra_labware should make that labware available."""
        explicit_extra_lw = {
            self.LW_LOAD_NAME: json.loads(
                load_shared_data(self.LW_FIXTURE_DIR / f"{self.LW_LOAD_NAME}.json")
            )
        }
        context = execute.get_protocol_api(api_version, extra_labware=explicit_extra_lw)
        assert context.load_labware(
            load_name=self.LW_LOAD_NAME, location=1, namespace=self.LW_NAMESPACE
        )

    def test_jupyter(
        self, api_version: APIVersion, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Putting labware in the Jupyter directory should make it available."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util,
            "JUPYTER_NOTEBOOK_LABWARE_DIR",
            get_shared_data_root() / self.LW_FIXTURE_DIR,
        )
        context = execute.get_protocol_api(api_version)
        assert context.load_labware(
            load_name=self.LW_LOAD_NAME, location=1, namespace=self.LW_NAMESPACE
        )

    def test_jupyter_override(
        self, api_version: APIVersion, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Passing any extra_labware should prevent searching the Jupyter directory."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util,
            "JUPYTER_NOTEBOOK_LABWARE_DIR",
            get_shared_data_root() / self.LW_FIXTURE_DIR,
        )
        context = execute.get_protocol_api(api_version, extra_labware={})
        with pytest.raises(Exception, match="Labware .+ not found"):
            context.load_labware(
                load_name=self.LW_LOAD_NAME, location=1, namespace=self.LW_NAMESPACE
            )

    def test_jupyter_not_on_filesystem(
        self, api_version: APIVersion, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """It should tolerate the Jupyter labware directory not existing on the filesystem."""
        # TODO(mm, 2023-10-06): This is monkeypatching a dependency of a dependency,
        # which is too deep.
        monkeypatch.setattr(entrypoint_util, "IS_ROBOT", True)
        monkeypatch.setattr(
            entrypoint_util, "JUPYTER_NOTEBOOK_LABWARE_DIR", HERE / "nosuchdirectory"
        )
        with_nonexistent_jupyter_extra_labware = execute.get_protocol_api(api_version)
        with pytest.raises(Exception, match="Labware .+ not found"):
            with_nonexistent_jupyter_extra_labware.load_labware(
                load_name=self.LW_LOAD_NAME, location=1, namespace=self.LW_NAMESPACE
            )



"""ProtocolEngine shared test fixtures."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from decoy import Decoy

from opentrons_shared_data import load_shared_data
from opentrons_shared_data.deck import load as load_deck
from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.labware import load_definition
from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition,
    labware_definition_type_adapter,
)
from opentrons_shared_data.pipette import pipette_definition

from opentrons.hardware_control import HardwareControlAPI, OT2HardwareControlAPI
from opentrons.hardware_control.api import API
from opentrons.hardware_control.protocols.types import FlexRobotType, OT2RobotType
from opentrons.protocol_engine.notes import CommandNoteAdder
from opentrons.protocol_engine.resources import CameraProvider, FileProvider
from opentrons.protocol_engine.types import ModuleDefinition
from opentrons.protocols.api_support.deck_type import (
    SHORT_TRASH_DECK,
    STANDARD_OT2_DECK,
    STANDARD_OT3_DECK,
)

if TYPE_CHECKING:
    from opentrons.hardware_control.ot3api import OT3API


@pytest.fixture
def hardware_api(decoy: Decoy) -> HardwareControlAPI:
    """Get a mocked out HardwareControlAPI of unspecified robot type."""
    return decoy.mock(cls=OT2HardwareControlAPI)


@pytest.fixture
def ot2_hardware_api(decoy: Decoy) -> API:
    """Get a mocked out OT-2 hardware API."""
    mock = decoy.mock(cls=API)
    decoy.when(mock.get_robot_type()).then_return(OT2RobotType)
    return mock


@pytest.fixture
def ot3_hardware_api(decoy: Decoy) -> OT3API:
    """Get a mocked out OT3API."""
    try:
        from opentrons.hardware_control.ot3api import OT3API

        mock = decoy.mock(cls=OT3API)
        decoy.when(mock.get_robot_type()).then_return(FlexRobotType)
        return mock
    except ImportError:
        # TODO (tz, 9-23-22) Figure out a better way to use this fixture with OT-3 api only.
        return None  # type: ignore[return-value]


@pytest.fixture(scope="session")
def ot2_standard_deck_def() -> DeckDefinitionV5:
    """Get the OT-2 standard deck definition."""
    return load_deck(STANDARD_OT2_DECK, 5)


@pytest.fixture(scope="session")
def ot2_short_trash_deck_def() -> DeckDefinitionV5:
    """Get the OT-2 short-trash deck definition."""
    return load_deck(SHORT_TRASH_DECK, 5)


@pytest.fixture(scope="session")
def ot3_standard_deck_def() -> DeckDefinitionV5:
    """Get the OT-2 standard deck definition."""
    return load_deck(STANDARD_OT3_DECK, 5)


@pytest.fixture(scope="session")
def ot2_fixed_trash_def() -> LabwareDefinition:
    """Get the definition of the OT-2 standard fixed trash."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_1_trash_1100ml_fixed", 1)
    )


@pytest.fixture(scope="session")
def ot2_short_fixed_trash_def() -> LabwareDefinition:
    """Get the definition of the OT-2 short fixed trash."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_1_trash_850ml_fixed", 1)
    )


@pytest.fixture(scope="session")
def ot3_fixed_trash_def() -> LabwareDefinition:
    """Get the definition of the OT-3 fixed trash."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_1_trash_3200ml_fixed", 1)
    )


@pytest.fixture(scope="session")
def ot3_absorbance_reader_lid() -> LabwareDefinition:
    """Get the definition of the OT-3 plate reader lid."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_flex_lid_absorbance_plate_reader_module", 1)
    )


@pytest.fixture(scope="session")
def well_plate_def() -> LabwareDefinition:
    """Get the definition of a 96 well plate."""
    return labware_definition_type_adapter.validate_python(
        load_definition("corning_96_wellplate_360ul_flat", 2)
    )


@pytest.fixture(scope="session")
def flex_50uL_tiprack() -> LabwareDefinition:
    """Get the definition of a Flex 50uL tiprack."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_flex_96_filtertiprack_50ul", 1)
    )


@pytest.fixture(scope="session")
def adapter_plate_def() -> LabwareDefinition:
    """Get the definition of a h/s adapter plate."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_universal_flat_adapter", 1)
    )


@pytest.fixture(scope="session")
def reservoir_def() -> LabwareDefinition:
    """Get the definition of single-row reservoir."""
    return labware_definition_type_adapter.validate_python(
        load_definition("nest_12_reservoir_15ml", 1)
    )


@pytest.fixture(scope="session")
def tip_rack_def() -> LabwareDefinition:
    """Get the definition of Opentrons 300 uL tip rack."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_96_tiprack_300ul", 1)
    )


@pytest.fixture(scope="session")
def tiprack_adapter_def() -> LabwareDefinition:
    """Get the definition of a flex 96 channel tiprack adapter."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_flex_96_tiprack_adapter", 1)
    )


@pytest.fixture(scope="session")
def adapter_def() -> LabwareDefinition:
    """Get the definition of Opentrons 96 PCR adapter."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_96_pcr_adapter", 1)
    )


@pytest.fixture(scope="session")
def lid_stack_def() -> LabwareDefinition:
    """Get the definition of the opentrons tiprack lid."""
    return labware_definition_type_adapter.validate_python(
        load_definition("protocol_engine_lid_stack_object", 1, schema=2)
    )


@pytest.fixture(scope="session")
def falcon_tuberack_def() -> LabwareDefinition:
    """Get the definition of the 6-well Falcon tuberack."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_6_tuberack_falcon_50ml_conical", 1)
    )


@pytest.fixture(scope="session")
def magdeck_well_plate_def() -> LabwareDefinition:
    """Get the definition of a well place compatible with magdeck."""
    return labware_definition_type_adapter.validate_python(
        load_definition("nest_96_wellplate_100ul_pcr_full_skirt", 1)
    )


@pytest.fixture(scope="session")
def tiprack_lid_def() -> LabwareDefinition:
    """Get the definition of the opentrons tiprack lid."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_flex_tiprack_lid", 1, schema=2)
    )


@pytest.fixture(scope="session")
def auto_sealing_lid_def() -> LabwareDefinition:
    """Get the definition of the opentrons tough auto sealing lid."""
    return labware_definition_type_adapter.validate_python(
        load_definition("opentrons_tough_pcr_auto_sealing_lid", 2, schema=2)
    )


@pytest.fixture(scope="session")
def tempdeck_v1_def() -> ModuleDefinition:
    """Get the definition of a V1 tempdeck."""
    definition = load_shared_data("module/definitions/3/temperatureModuleV1.json")
    return ModuleDefinition.model_validate_json(definition)


@pytest.fixture(scope="session")
def tempdeck_v2_def() -> ModuleDefinition:
    """Get the definition of a V2 tempdeck."""
    definition = load_shared_data("module/definitions/3/temperatureModuleV2.json")
    return ModuleDefinition.model_validate_json(definition)


@pytest.fixture(scope="session")
def magdeck_v1_def() -> ModuleDefinition:
    """Get the definition of a V1 magdeck."""
    definition = load_shared_data("module/definitions/3/magneticModuleV1.json")
    return ModuleDefinition.model_validate_json(definition)


@pytest.fixture(scope="session")
def magdeck_v2_def() -> ModuleDefinition:
    """Get the definition of a V2 magdeck."""
    definition = load_shared_data("module/definitions/3/magneticModuleV2.json")
    return ModuleDefinition.model_validate_json(definition)


@pytest.fixture(scope="session")
def thermocycler_v1_def() -> ModuleDefinition:
    """Get the definition of a V2 thermocycler."""
    definition = load_shared_data("module/definitions/3/thermocyclerModuleV1.json")
    return ModuleDefinition.model_validate_json(definition)


@pytest.fixture(scope="session")
def thermocycler_v2_def() -> ModuleDefinition:
    """Get the definition of a V2 thermocycler."""
    definition = load_shared_data("module/definitions/3/thermocyclerModuleV2.json")
    return ModuleDefinition.model_validate_json(definition)


@pytest.fixture(scope="session")
def heater_shaker_v1_def() -> ModuleDefinition:
    """Get the definition of a V1 heater-shaker."""
    definition = load_shared_data("module/definitions/3/heaterShakerModuleV1.json")
    return ModuleDefinition.model_validate_json(definition)


@pytest.fixture(scope="session")
def mag_block_v1_def() -> ModuleDefinition:
    """Get the definition of a V1 Mag Block."""
    definition = load_shared_data("module/definitions/3/magneticBlockV1.json")
    return ModuleDefinition.model_validate_json(definition)


@pytest.fixture(scope="session")
def abs_reader_v1_def() -> ModuleDefinition:
    """Get the definition of a V1 absorbance plate reader."""
    definition = load_shared_data("module/definitions/3/absorbanceReaderV1.json")
    return ModuleDefinition.model_validate_json(definition)


@pytest.fixture(scope="session")
def flex_stacker_v1_def() -> ModuleDefinition:
    """Get the definition of a V1 Flex Stacker."""
    definition = load_shared_data("module/definitions/3/flexStackerModuleV1.json")
    return ModuleDefinition.model_validate_json(definition)


@pytest.fixture(scope="session")
def supported_tip_fixture() -> pipette_definition.SupportedTipsDefinition:
    """Get a mock supported tip definition."""
    return pipette_definition.SupportedTipsDefinition(
        defaultAspirateFlowRate=pipette_definition.FlowRateDefinition(
            default=10, valuesByApiLevel={}
        ),
        defaultDispenseFlowRate=pipette_definition.FlowRateDefinition(
            default=10, valuesByApiLevel={}
        ),
        defaultBlowOutFlowRate=pipette_definition.FlowRateDefinition(
            default=10, valuesByApiLevel={}
        ),
        defaultTipLength=40,
        defaultReturnTipHeight=0.5,
        aspirate=pipette_definition.ulPerMMDefinition(default={"1": [(0, 0, 0)]}),
        dispense=pipette_definition.ulPerMMDefinition(default={"1": [(0, 0, 0)]}),
        defaultPushOutVolume=3,
    )


@pytest.fixture
def mock_command_note_adder(decoy: Decoy) -> CommandNoteAdder:
    """Get a command note adder."""
    return decoy.mock(cls=CommandNoteAdder)


@pytest.fixture
def file_provider(decoy: Decoy) -> FileProvider:
    """Get a mocked out FileProvider."""
    return decoy.mock(cls=FileProvider)


@pytest.fixture
def camera_provider(decoy: Decoy) -> CameraProvider:
    """Get a mocked out CameraProvider."""
    return decoy.mock(cls=CameraProvider)



from __future__ import annotations

import asyncio
import logging
from typing import AsyncIterator, Literal, Type

from .async_serial import AsyncSerial
from .errors import (
    AlarmResponse,
    BaseErrorCode,
    DefaultErrorCodes,
    ErrorResponse,
    GCodeCacheFull,
    NoResponse,
    UnhandledGcode,
)
from opentrons.drivers.command_builder import CommandBuilder

log = logging.getLogger(__name__)


class SerialConnection:
    @classmethod
    async def _build_serial(
        cls,
        port: str,
        baud_rate: int,
        timeout: float,
        loop: asyncio.AbstractEventLoop | None,
        reset_buffer_before_write: bool,
    ) -> AsyncSerial:
        return await AsyncSerial.create(
            port=port,
            baud_rate=baud_rate,
            timeout=timeout,
            loop=loop,
            reset_buffer_before_write=reset_buffer_before_write,
        )

    @classmethod
    async def create(
        cls,
        port: str,
        baud_rate: int,
        timeout: float,
        ack: str,
        name: str | None = None,
        retry_wait_time_seconds: float = 0.1,
        loop: asyncio.AbstractEventLoop | None = None,
        error_keyword: str | None = None,
        alarm_keyword: str | None = None,
        reset_buffer_before_write: bool = False,
        error_codes: Type[BaseErrorCode] = DefaultErrorCodes,
    ) -> "SerialConnection":
        """
        Create a connection.

        Args:
            port: url or port to connect to
            baud_rate: baud rate
            timeout: timeout in seconds
            ack: the command response ack
            name: the connection name
            retry_wait_time_seconds: how long to wait between retries.
            loop: optional event loop.
            error_keyword: optional string that will cause an
                           ErrorResponse exception when detected
                           (default: error)
            alarm_keyword: optional string that will cause an
                           AlarmResponse exception when detected
                           (default: alarm)
            reset_buffer_before_write: whether to reset the read buffer before
              every write
            error_codes: Enum class for error codes
                         (default: DefaultErrorCodes)

        Returns: SerialConnection
        """
        serial = await cls._build_serial(
            port=port,
            baud_rate=baud_rate,
            timeout=timeout,
            loop=loop,
            reset_buffer_before_write=reset_buffer_before_write,
        )
        name = name or port
        obj = cls(
            serial=serial,
            port=port,
            name=name,
            ack=ack,
            retry_wait_time_seconds=retry_wait_time_seconds,
            error_keyword=error_keyword or "error",
            alarm_keyword=alarm_keyword or "alarm",
            error_codes=error_codes,
        )
        await obj.flush_input()
        return obj

    def __init__(
        self,
        serial: AsyncSerial,
        port: str,
        name: str,
        ack: str,
        retry_wait_time_seconds: float,
        error_keyword: str,
        alarm_keyword: str,
        error_codes: Type[BaseErrorCode] = DefaultErrorCodes,
    ) -> None:
        """
        Constructor

        Args:
            serial: AsyncSerial object
            port: url or port to connect to
            ack: the command response ack
            name: the connection name
            retry_wait_time_seconds: how long to wait between retries.
            error_keyword: string that will cause an ErrorResponse
                           exception when detected
            alarm_keyword: string that will cause an AlarmResponse
                           exception when detected
            error_codes: Enum class for error codes
        """
        self._serial = serial
        self._port = port
        self._name = name
        self._ack = ack.encode()
        self._retry_wait_time_seconds = retry_wait_time_seconds
        self._send_data_lock = asyncio.Lock()
        self._error_keyword = error_keyword.lower()
        self._alarm_keyword = alarm_keyword.lower()
        self._error_codes = error_codes

    async def send_command(
        self, command: CommandBuilder, retries: int = 0, timeout: float | None = None
    ) -> str:
        """
        Send a command and return the response.

        Args:
            command: A command builder.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds

        Returns: The command response

        Raises: SerialException
        """
        return await self.send_data(
            data=command.build(), retries=retries, timeout=timeout
        )

    async def send_dfu_command(self, command: CommandBuilder) -> None:
        """
        Send a dfu command to enter device bootloader.

        This method doesn't wait for a response after sending the command since the
        module port gets disconnected once it enters its bootloader.
        """
        encoded_command = command.build().encode()

        async with self._send_data_lock:
            log.debug(f"{self.name}: Write -> {encoded_command!r}")
            await self._serial.write(data=encoded_command)

    async def send_data(
        self, data: str, retries: int = 0, timeout: float | None = None
    ) -> str:
        """
        Send data and return the response.

        Args:
            data: The data to send.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds

        Returns: The command response

        Raises: SerialException
        """
        async with (
            self._send_data_lock,
            self._serial.timeout_override("timeout", timeout),
        ):
            return await self._send_data(data=data, retries=retries)

    async def _send_data(self, data: str, retries: int) -> str:
        """
        Send data and return the response.

        Args:
            data: The data to send.
            retries: number of times to retry in case of timeout

        Returns: The command response

        Raises: SerialException
        """
        data_encode = data.encode()

        for retry in range(retries + 1):
            log.debug(f"{self.name}: Write -> {data_encode!r}")
            await self._serial.write(data=data_encode)

            response = await self._serial.read_until(match=self._ack)
            log.debug(f"{self.name}: Read <- {response!r}")

            if (
                self._ack in response
                or self._error_keyword.encode() in response.lower()
            ):
                # Remove ack from response
                response = response.replace(self._ack, b"")
                str_response = self.process_raw_response(
                    command=data, response=response.decode()
                )
                self.raise_on_error(response=str_response, request=data)
                return str_response

            log.info(f"{self.name}: retry number {retry}/{retries}")

            await self.on_retry()

        raise NoResponse(port=self._port, command=data)

    async def open(self) -> None:
        """Open the connection."""
        await self._serial.open()

    async def close(self) -> None:
        """Close the connection."""
        await self._serial.close()

    async def is_open(self) -> bool:
        """Check if connection is open."""
        return await self._serial.is_open()

    @property
    def port(self) -> str:
        return self._port

    @property
    def name(self) -> str:
        return self._name

    @property
    def send_data_lock(self) -> asyncio.Lock:
        return self._send_data_lock

    def raise_on_error(self, response: str, request: str) -> None:
        """
        Raise an error if the response contains an error

        Args:
            response: response
            request: the requesting command

        Returns: None

        Raises: SerialException
        """
        if not response or not request:
            return

        lower = response.lower()
        try:
            res_gcode = response.split()[0]
            req_gcode = request.split()[0]
        except IndexError:
            # this means the response is an empty string or something, which is weird
            # but not a canonical error
            return

        # Make sure this is not just a normal response that happens to contain the
        # `err` or `alarm` keyword in the message body by checking the gcode values
        # for both the request and response. If the gcodes are the same then this
        # is not an error response.
        if res_gcode == req_gcode:
            return

        if self._alarm_keyword in lower:
            raise AlarmResponse(port=self._port, response=response)

        if self._error_keyword.lower() in lower:
            # Check for specific error codes
            error_codes_dict = self._error_codes.get_error_codes()
            for code, error_code in error_codes_dict.items():
                if code in lower:
                    error_code.raise_exception(
                        port=self._port, response=response, command=request
                    )

            # If no specific error code was found, raise a generic ErrorResponse
            raise ErrorResponse(port=self._port, response=response)

    async def on_retry(self) -> None:
        """
        Opportunity for derived classes to perform action between retries. Default
        behaviour is to wait then re-open the connection.

        Returns: None
        """
        await asyncio.sleep(self._retry_wait_time_seconds)
        await self._serial.close()
        await self._serial.open()

    def process_raw_response(self, command: str, response: str) -> str:
        """
        Opportunity for derived classes to process the raw response. Default
         strips white space.

        Args:
            command: The sent command.
            response: The raw read response minus ack.

        Returns:
            processed response.
        """
        return response.strip()

    async def flush_input(self) -> None:
        """Empty the input buffer.

        This is a pretty gross utility that may take a while and is intended to consume
        blocks of text printed by the other side, for instance on boot.
        """
        self._serial.reset_input_buffer()
        log.info("flushing input")
        consecutive_empties = 0
        async with self._serial.timeout_override("timeout", 0.1):
            while True:
                try:
                    inp = await self._serial.read_until(b"\r\n")
                    log.info(f"flush_input read: {inp!r}")
                    if not inp:
                        consecutive_empties += 1
                        if consecutive_empties >= 5:
                            return
                    else:
                        consecutive_empties = 0
                except Exception:
                    log.exception("timeout exception is")
                    return


class AsyncResponseSerialConnection(SerialConnection):
    @classmethod
    async def create(
        cls,
        port: str,
        baud_rate: int,
        timeout: float,
        ack: str,
        name: str | None = None,
        retry_wait_time_seconds: float = 0.1,
        loop: asyncio.AbstractEventLoop | None = None,
        error_keyword: str | None = None,
        alarm_keyword: str | None = None,
        reset_buffer_before_write: bool = False,
        error_codes: Type[BaseErrorCode] = DefaultErrorCodes,
        async_error_ack: str | None = None,
        number_of_retries: int = 0,
    ) -> AsyncResponseSerialConnection:
        """
        Create a connection.

        Args:
            port: url or port to connect to
            baud_rate: baud rate
            timeout: timeout in seconds
            ack: the command response ack
            name: the connection name
            retry_wait_time_seconds: how long to wait between retries.
            loop: optional event loop.
            error_keyword: optional string that will cause an
                           ErrorResponse exception when detected
                           (default: error)
            alarm_keyword: optional string that will cause an
                           AlarmResponse exception when detected
                           (default: alarm)
            reset_buffer_before_write: whether to reset the read buffer before
              every write
            async_error_ack: optional string that will indicate an asynchronous
                             error when detected (default: async)
            number_of_retries: default number of retries
            error_codes: Enum class for error codes
                         (default: DefaultErrorCodes)

        Returns: AsyncResponseSerialConnection
        """
        serial = await super()._build_serial(
            port=port,
            baud_rate=baud_rate,
            timeout=timeout,
            loop=loop,
            reset_buffer_before_write=reset_buffer_before_write,
        )
        name = name or port
        obj = cls(
            serial=serial,
            port=port,
            name=name,
            ack=ack,
            retry_wait_time_seconds=retry_wait_time_seconds,
            error_keyword=error_keyword or "err",
            alarm_keyword=alarm_keyword or "alarm",
            async_error_ack=async_error_ack or "async",
            number_of_retries=number_of_retries,
            error_codes=error_codes,
        )
        return obj

    def __init__(
        self,
        serial: AsyncSerial,
        port: str,
        name: str,
        ack: str,
        retry_wait_time_seconds: float,
        error_keyword: str,
        alarm_keyword: str,
        async_error_ack: str,
        number_of_retries: int = 0,
        error_codes: Type[BaseErrorCode] = DefaultErrorCodes,
    ) -> None:
        """
        Constructor

        Args:
            serial: AsyncSerial object
            port: url or port to connect to
            ack: the command response ack
            name: the connection name
            retry_wait_time_seconds: how long to wait between retries.
            error_keyword: string that will cause an ErrorResponse
                           exception when detected
            alarm_keyword: string that will cause an AlarmResponse
                           exception when detected
            async_error_ack: string that will indicate an asynchronous
                             error when detected
            number_of_retries: default number of retries
            error_codes: Enum class for error codes
        """
        super().__init__(
            serial=serial,
            port=port,
            name=name,
            ack=ack,
            retry_wait_time_seconds=retry_wait_time_seconds,
            error_keyword=error_keyword,
            alarm_keyword=alarm_keyword,
            error_codes=error_codes,
        )
        self._serial = serial
        self._port = port
        self._name = name
        self._ack = ack.encode()
        self._retry_wait_time_seconds = retry_wait_time_seconds
        self._number_of_retries = number_of_retries
        self._error_keyword = error_keyword.lower()
        self._alarm_keyword = alarm_keyword.lower()
        self._async_error_ack = async_error_ack.lower()

    async def send_multiack_command(
        self,
        command: CommandBuilder,
        retries: int = 0,
        timeout: float | None = None,
        acks: int = 1,
    ) -> list[str]:
        """Send a command and return the responses.

        Some commands result in multiple responses; collate them and return them all.

        Args:
            command: A command builder.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds
            acks: the number of acks to expect
        """
        return await self.send_data_multiack(
            data=command.build(), retries=retries, timeout=timeout, acks=acks
        )

    async def send_command(
        self,
        command: CommandBuilder,
        retries: int | None = None,
        timeout: float | None = None,
    ) -> str:
        """
        Send a command and return the response.

        Args:
            command: A command builder.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds

        Returns: The command response

        Raises: SerialException
        """
        return await self.send_data(
            data=command.build(),
            retries=retries if retries is not None else self._number_of_retries,
            timeout=timeout,
        )

    async def send_data_multiack(
        self, data: str, retries: int = 0, timeout: float | None = None, acks: int = 1
    ) -> list[str]:
        """Send data and return all responses."""
        async with (
            super().send_data_lock,
            self._serial.timeout_override("timeout", timeout),
        ):
            return await self._send_data_multiack(
                data=data, retries=retries or self._number_of_retries, acks=acks
            )

    async def send_data(
        self, data: str, retries: int | None = None, timeout: float | None = None
    ) -> str:
        """
        Send data and return the response.

        Args:
            data: The data to send.
            retries: number of times to retry in case of timeout
            timeout: optional override of default timeout in seconds

        Returns: The command response

        Raises: SerialException
        """
        async with (
            super().send_data_lock,
            self._serial.timeout_override("timeout", timeout),
        ):
            return await self._send_data(
                data=data,
                retries=retries if retries is not None else self._number_of_retries,
            )

    async def _consume_responses(
        self, acks: int
    ) -> AsyncIterator[tuple[Literal["response", "error", "empty-unknown"], bytes]]:
        while acks > 0:
            data = await self._serial.read_until(match=self._ack)
            log.debug(f"{self._name}: Read <- {data!r}")
            if self._async_error_ack.encode() in data:
                yield "error", data
            elif self._ack in data:
                yield "response", data
                acks -= 1
            else:
                # A read timeout, end
                yield "empty-unknown", data

    def _raise_on_parser_error(self, data: str, response: bytes) -> None:
        """Raise an exception if this response contains an error from the gcode parser on the module.

        This has to be treated specially because multiack commands won't get multiple acks if the command
        fails at the parse stage. The errors handled here should be kept in sync with the module gcode
        parse code.
        """
        try:
            str_response = self.process_raw_response(
                command=data, response=response.replace(self._ack, b"").decode()
            )
            self.raise_on_error(response=str_response, request=data)
        except (UnhandledGcode, GCodeCacheFull):
            raise
        except Exception:
            pass

    async def _send_one_retry(self, data: str, acks: int) -> list[str]:
        data_encode = data.encode("utf-8")
        log.debug(f"{self._name}: Write -> {data_encode!r}")
        await self._serial.write(data=data_encode)

        command_acks: list[bytes] = []
        async_errors: list[bytes] = []
        # consume responses before raising so we don't raise and orphan
        # a response in the buffer
        async for response_type, response in self._consume_responses(acks):
            if response_type == "error":
                async_errors.append(response)
                self._raise_on_parser_error(data, response)
            elif response_type == "response":
                command_acks.append(response)
                self._raise_on_parser_error(data, response)
            else:
                break

        for async_error in async_errors:
            # Remove ack from response
            ackless_response = async_error.replace(self._ack, b"")
            str_response = self.process_raw_response(
                command=data, response=ackless_response.decode()
            )
            self.raise_on_error(response=str_response, request=data)

        ackless_responses: list[str] = []
        for command_ack in command_acks:
            # Remove ack from response
            ackless_response = command_ack.replace(self._ack, b"")
            str_response = self.process_raw_response(
                command=data, response=ackless_response.decode()
            )
            self.raise_on_error(response=str_response, request=data)
            ackless_responses.append(str_response)
        return ackless_responses

    async def _send_data_multiack(
        self, data: str, retries: int, acks: int
    ) -> list[str]:
        """
        Send data and return the response(s).

        Args:
            data: The data to send.
            retries: number of times to retry in case of timeout
            acks: The number of expected command responses

        This function retries (resends the command) up to (retries) times, and waits
        for (acks) responses. It also listens for async errors. These are an older
        mechanism where at the moment an error occurs, some modules will send a message
        like async error ERR:202:whatever

        This function will detect async error messages if they were sent before it
        sent the command or if they are sent before the final ack for the command is
        sent. It will not catch async errors otherwise.

        This function will always try and consume all the acknowledgements specified for
        its command if it sends the command, even if an async error happens in between.

        This should all work together to make sure that there aren't any leftover acks
        after the function ends, which could lead to the read/write mechanics getting out
        of sync.

        Returns: The command responses

        Raises: SerialException from an error ack to this command or an async error.
        """
        responses: list[str] = []

        for retry in range(retries + 1):
            responses = await self._send_one_retry(data, acks)
            if responses:
                return responses
            log.info(f"{self._name}: retry number {retry}/{retries}")
            await self.on_retry()

        raise NoResponse(port=self._port, command=data)

    async def _send_data(self, data: str, retries: int) -> str:
        return (await self._send_data_multiack(data, retries, 1))[0]



from .abstract import AbstractVacuumModuleDriver
from .driver import VacuumModuleDriver
from .simulator import SimulatingDriver

__all__ = ["VacuumModuleDriver", "SimulatingDriver", "AbstractVacuumModuleDriver"]



"""Test setStatusBar commands."""

import pytest
from decoy import Decoy

from opentrons.hardware_control.types import StatusBarState
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.set_status_bar import (
    SetStatusBarImplementation,
    SetStatusBarParams,
    SetStatusBarResult,
    StatusBarAnimation,
)
from opentrons.protocol_engine.execution.status_bar import StatusBarHandler


@pytest.fixture
def subject(
    status_bar: StatusBarHandler,
) -> SetStatusBarImplementation:
    """Returns subject under test."""
    return SetStatusBarImplementation(status_bar=status_bar)


async def test_status_bar_busy(
    decoy: Decoy,
    status_bar: StatusBarHandler,
    subject: SetStatusBarImplementation,
) -> None:
    """Test when the status bar is busy."""
    decoy.when(status_bar.status_bar_should_not_be_changed()).then_return(True)

    data = SetStatusBarParams(animation=StatusBarAnimation.OFF)

    result = await subject.execute(params=data)

    assert result == SuccessData(public=SetStatusBarResult())

    decoy.verify(await status_bar.set_status_bar(status=StatusBarState.OFF), times=0)


@pytest.mark.parametrize(
    argnames=["animation", "expected_state"],
    argvalues=[
        [StatusBarAnimation.CONFIRM, StatusBarState.CONFIRMATION],
        [StatusBarAnimation.DISCO, StatusBarState.DISCO],
        [StatusBarAnimation.OFF, StatusBarState.OFF],
        [StatusBarAnimation.IDLE, StatusBarState.IDLE],
        [StatusBarAnimation.UPDATING, StatusBarState.UPDATING],
    ],
)
async def test_set_status_bar_animation(
    decoy: Decoy,
    status_bar: StatusBarHandler,
    subject: SetStatusBarImplementation,
    animation: StatusBarAnimation,
    expected_state: StatusBarState,
) -> None:
    """Test when status bar is NOT busy."""
    decoy.when(status_bar.status_bar_should_not_be_changed()).then_return(False)

    data = SetStatusBarParams(animation=animation)

    result = await subject.execute(params=data)
    assert result == SuccessData(public=SetStatusBarResult())

    decoy.verify(await status_bar.set_status_bar(status=expected_state), times=1)



"""Well state store tests.

DEPRECATED: Testing WellStore independently of WellView is no longer helpful.
Try to add new tests to well_state.py, where they can be tested together,
treating WellState as a private implementation detail.
"""

from datetime import datetime

import pytest

from .command_fixtures import (
    create_aspirate_command,
    create_liquid_probe_command,
    create_load_liquid_command,
)
from opentrons.protocol_engine.actions.actions import SucceedCommandAction
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.wells import WellStore


@pytest.fixture
def subject() -> WellStore:
    """Well store test subject."""
    return WellStore()


def test_handles_liquid_probe_success(subject: WellStore) -> None:
    """It should add the well to the state after a successful liquid probe."""
    labware_id = "labware-id"
    well_name = "well-name"
    liquid_probe = create_liquid_probe_command()
    timestamp = datetime(year=2020, month=1, day=2)

    subject.handle_action(
        SucceedCommandAction(
            command=liquid_probe,
            state_update=update_types.StateUpdate(
                liquid_probed=update_types.LiquidProbedUpdate(
                    labware_id="labware-id",
                    well_name="well-name",
                    height=15.0,
                    volume=30.0,
                    last_probed=timestamp,
                )
            ),
        )
    )

    assert len(subject.state.probed_heights) == 1
    assert len(subject.state.probed_volumes) == 1

    assert subject.state.probed_heights[labware_id][well_name].height == 15.0
    assert subject.state.probed_heights[labware_id][well_name].last_probed == timestamp
    assert subject.state.probed_volumes[labware_id][well_name].volume == 30.0
    assert subject.state.probed_volumes[labware_id][well_name].last_probed == timestamp
    assert (
        subject.state.probed_volumes[labware_id][well_name].operations_since_probe == 0
    )


def test_handles_load_liquid_success(subject: WellStore) -> None:
    """It should add the well to the state after a successful load liquid."""
    labware_id = "labware-id"
    well_name_1 = "well-name-1"
    well_name_2 = "well-name-2"
    load_liquid = create_load_liquid_command(
        labware_id=labware_id, volume_by_well={well_name_1: 30, well_name_2: 100}
    )
    timestamp = datetime(year=2020, month=1, day=2)

    subject.handle_action(
        SucceedCommandAction(
            command=load_liquid,
            state_update=update_types.StateUpdate(
                liquid_loaded=update_types.LiquidLoadedUpdate(
                    labware_id=labware_id,
                    volumes={well_name_1: 30, well_name_2: 100},
                    last_loaded=timestamp,
                )
            ),
        )
    )

    assert len(subject.state.loaded_volumes) == 1
    assert len(subject.state.loaded_volumes[labware_id]) == 2

    assert subject.state.loaded_volumes[labware_id][well_name_1].volume == 30.0
    assert (
        subject.state.loaded_volumes[labware_id][well_name_1].last_loaded == timestamp
    )
    assert (
        subject.state.loaded_volumes[labware_id][well_name_1].operations_since_load == 0
    )
    assert subject.state.loaded_volumes[labware_id][well_name_2].volume == 100.0
    assert (
        subject.state.loaded_volumes[labware_id][well_name_2].last_loaded == timestamp
    )
    assert (
        subject.state.loaded_volumes[labware_id][well_name_2].operations_since_load == 0
    )


def test_handles_load_liquid_and_aspirate(subject: WellStore) -> None:
    """It should populate the well state after load liquid and update the well state after aspirate."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name_1 = "well-name-1"
    well_name_2 = "well-name-2"
    aspirated_volume = 10.0
    load_liquid = create_load_liquid_command(
        labware_id=labware_id, volume_by_well={well_name_1: 30, well_name_2: 100}
    )
    aspirate_1 = create_aspirate_command(
        pipette_id=pipette_id,
        volume=aspirated_volume,
        flow_rate=1.0,
        labware_id=labware_id,
        well_name=well_name_1,
    )
    aspirate_2 = create_aspirate_command(
        pipette_id=pipette_id,
        volume=aspirated_volume,
        flow_rate=1.0,
        labware_id=labware_id,
        well_name=well_name_2,
    )
    timestamp = datetime(year=2020, month=1, day=2)

    subject.handle_action(
        SucceedCommandAction(
            command=load_liquid,
            state_update=update_types.StateUpdate(
                liquid_loaded=update_types.LiquidLoadedUpdate(
                    labware_id=labware_id,
                    volumes={well_name_1: 30, well_name_2: 100},
                    last_loaded=timestamp,
                )
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=aspirate_1,
            state_update=update_types.StateUpdate(
                liquid_operated=update_types.LiquidOperatedUpdate(
                    labware_id=labware_id,
                    well_names=[well_name_1, well_name_2],
                    volume_added=-aspirated_volume,
                )
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=aspirate_2,
            state_update=update_types.StateUpdate(
                liquid_operated=update_types.LiquidOperatedUpdate(
                    labware_id=labware_id,
                    well_names=[well_name_2],
                    volume_added=-aspirated_volume,
                )
            ),
        )
    )

    assert len(subject.state.loaded_volumes) == 1
    assert len(subject.state.loaded_volumes[labware_id]) == 2

    assert subject.state.loaded_volumes[labware_id][well_name_1].volume == 20.0
    assert (
        subject.state.loaded_volumes[labware_id][well_name_1].last_loaded == timestamp
    )
    assert (
        subject.state.loaded_volumes[labware_id][well_name_1].operations_since_load == 1
    )
    assert subject.state.loaded_volumes[labware_id][well_name_2].volume == 80.0
    assert (
        subject.state.loaded_volumes[labware_id][well_name_2].last_loaded == timestamp
    )
    assert (
        subject.state.loaded_volumes[labware_id][well_name_2].operations_since_load == 2
    )


def test_handles_liquid_probe_and_aspirate(subject: WellStore) -> None:
    """It should populate the well state after liquid probe and update the well state after aspirate."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    aspirated_volume = 10.0
    liquid_probe = create_liquid_probe_command()
    aspirate = create_aspirate_command(
        pipette_id=pipette_id,
        volume=aspirated_volume,
        flow_rate=1.0,
        labware_id=labware_id,
        well_name=well_name,
    )
    timestamp = datetime(year=2020, month=1, day=2)

    subject.handle_action(
        SucceedCommandAction(
            command=liquid_probe,
            state_update=update_types.StateUpdate(
                liquid_probed=update_types.LiquidProbedUpdate(
                    labware_id="labware-id",
                    well_name="well-name",
                    height=15.0,
                    volume=30.0,
                    last_probed=timestamp,
                )
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=aspirate,
            state_update=update_types.StateUpdate(
                liquid_operated=update_types.LiquidOperatedUpdate(
                    labware_id="labware-id",
                    well_names=["well-name"],
                    volume_added=-aspirated_volume,
                )
            ),
        )
    )

    assert len(subject.state.probed_heights[labware_id]) == 0
    assert len(subject.state.probed_volumes) == 1

    assert subject.state.probed_volumes[labware_id][well_name].volume == 20.0
    assert subject.state.probed_volumes[labware_id][well_name].last_probed == timestamp
    assert (
        subject.state.probed_volumes[labware_id][well_name].operations_since_probe == 1
    )



"""Test Flex Stacker prepare shuttle command implementation."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import FlexStackerStallError

from opentrons.drivers.flex_stacker.types import StackerAxis
from opentrons.hardware_control.modules import FlexStacker
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.flex_stacker.common import (
    FlexStackerStallOrCollisionError,
)
from opentrons.protocol_engine.commands.unsafe.unsafe_stacker_prepare_shuttle import (
    UnsafeFlexStackerPrepareShuttleImpl,
    UnsafeFlexStackerPrepareShuttleParams,
    UnsafeFlexStackerPrepareShuttleResult,
)
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.module_substates import (
    FlexStackerId,
    FlexStackerSubState,
)
from opentrons.protocol_engine.state.state import StateView


@pytest.fixture
def subject(
    state_view: StateView, equipment: EquipmentHandler, model_utils: ModelUtils
) -> UnsafeFlexStackerPrepareShuttleImpl:
    """Get a UnsafeFlexStackerPrepareShuttle command to test."""
    return UnsafeFlexStackerPrepareShuttleImpl(
        state_view=state_view, equipment=equipment, model_utils=model_utils
    )


async def test_home_command(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: UnsafeFlexStackerPrepareShuttleImpl,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
) -> None:
    """It should return a success data."""
    data = UnsafeFlexStackerPrepareShuttleParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=None,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=[],
        max_pool_count=0,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    result = await subject.execute(data)

    decoy.verify(await stacker_hardware.home_all(False), times=1)

    assert result == SuccessData(public=UnsafeFlexStackerPrepareShuttleResult())


async def test_home_command_with_stall_detected(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: UnsafeFlexStackerPrepareShuttleImpl,
    model_utils: ModelUtils,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
) -> None:
    """It should return a success data."""
    err_id = "error-id"
    err_timestamp = datetime(year=2025, month=3, day=19)

    data = UnsafeFlexStackerPrepareShuttleParams(moduleId=stacker_id)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=None,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=[],
        max_pool_count=0,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)
    decoy.when(model_utils.generate_id()).then_return(err_id)
    decoy.when(model_utils.get_timestamp()).then_return(err_timestamp)

    decoy.when(await stacker_hardware.home_all(False)).then_raise(  # type: ignore[func-returns-value]
        FlexStackerStallError(serial="123", axis=StackerAxis.Z)
    )

    result = await subject.execute(data)

    assert result == DefinedErrorData(
        public=FlexStackerStallOrCollisionError.model_construct(
            id=err_id,
            createdAt=err_timestamp,
            wrappedErrors=[matchers.Anything()],
            errorInfo={},
        )
    )
